### Least Ambiguous Set-Valued Classifiers with Bounded Error Levels
### Example 5.4: Abalone dataset

rm(list = ls())
library(glmnet)

#######################################################
## Conformal LABEL thresholds for total coverage and class-specific coverage
#######################################################
conf_thlds <- function(phat,Y2,alpha_tot,class_alphas){
     m = nrow(phat)
	 K = ncol(phat)
     score = rep(0,m)
     for(i in 1:m){
          score[i] = phat[i,Y2[i]]
     }
	 ## Total coverage
	 tot_thlds = sort(score)[ceiling(alpha_tot*(m+1)-1)]
	 
	 ## Class-specific coverage
	 class_thlds <- rep(NA,K)
	 for (k in 1:K){
		class_thlds[k] <- sort(score[Y2==k])[ceiling(class_alphas[k]*(sum(Y2==k)+1)-1)]
	 }
	 
	 return(list(tot_thlds=tot_thlds,class_thlds=class_thlds))
}

### Abalone data
set.seed(2017)
library(AppliedPredictiveModeling)
data(abalone)
Y = abalone[,9]
X <- as.matrix(cbind(model.matrix(~abalone[,1]-1), abalone[,2:8]))
n = length(Y)
Z = rep(0,n)
Z[Y <= 8] = 1
Z[(Y > 8)&(Y < 11)] = 2
Z[Y >= 11] = 3
Y = Z

# Split into testing and training
n.test <- floor(n/5)
rand.ind <- sample.int(n)
test.ind <- rand.ind[1:n.test]
train.ind <- rand.ind[(n.test+1):n]

X.train <- X[train.ind,]
Y.train <- Y[train.ind]
X.test <- X[test.ind,]
Y.test <- Y[test.ind]

# Split the training data for split-conformal prediction
n.train = length(Y.train)
m = round(n.train/2)
I = sample(1:n.train, size=m, replace=FALSE)
# model fitting data
Y1 = Y.train[I]
X1 = X.train[I,]
n1 = length(Y1)
# threshold estimation data
J = setdiff(1:n.train,I)
Y2 = Y.train[J]
X2 = X.train[J,]
n2 = length(Y2)

# fit model on (X1,Y1)
lambda <- cv.glmnet(X1, Y1, family="multinomial")$lambda.min
fitted.model <- glmnet(X1, Y1, family="multinomial", lambda = lambda)
# evaluate model on X2
phat <- predict(fitted.model, newx = X2, type = "response")[,,1]
# get thresholds for total and class specific coverage
(thlds_est <- conf_thlds(phat,Y2,alpha_tot=.1,class_alphas=rep(.1,3)))

# Evaluate ambiguity and coverage on test data
# evaluate model on X.test
phat_test <- predict(fitted.model, newx = X.test, type = "response")[,,1]

Hstar_totcov <- t(apply(phat_test, 1, 
		function(x){
			as.numeric(x >= rep(thlds_est$tot_thlds,3))
		}))	
mean(rowSums(Hstar_totcov)) # ambiguity 

# test sample class coverage of classifier that controls for total coverage
round( sapply(1:3,function(k)mean(Hstar_totcov[Y.test==k,k])), 3)


Hstar_classcov <- t(apply(phat_test, 1, 
		function(x){
			as.numeric(x >= thlds_est$class_thlds)
		}))	
mean(rowSums(Hstar_classcov)) # ambiguity

# Illustration with Male subset of abalones
test.subset.ind <- which(X.test[,3]==1)
Hstar_classcov_subset <- Hstar_classcov[test.subset.ind,]

ambi12 <- which(rowSums(Hstar_classcov_subset[,1:2])>1)
ambi23 <- which(rowSums(Hstar_classcov_subset[,2:3])>1)

X.pca <- princomp(X.test[test.subset.ind,-c(1:3)])
Y.test.subset <- Y.test[test.subset.ind]
Y.color <- rep('red',length(Y.test.subset))
Y.color[which(Y.test.subset==2)] <- 'gold'
Y.color[which(Y.test.subset==3)] <- 'blue'

windows(8,6)
par(mar = c(4.5,4.5,3,1))
plot(-X.pca$score[,1], X.pca$score[,2],col=Y.color,pch=as.character(Y.test.subset),xlab='Principal Component 1',ylab='Principal Component 2',
	cex.axis=1.5,cex.lab=1.8,lwd=3,cex=1.5,main="Abalone Data",cex.main=2)
points(-X.pca$score[ambi23,1],X.pca$score[ambi23,2],col='forestgreen',pch=0,lwd=2,cex=2.5)
points(-X.pca$score[ambi12,1],X.pca$score[ambi12,2],col='orange',pch=1,lwd=2,cex=2.5)


#######################################################
## Comparing to classification with rejection (CWR)
#######################################################

CWR <- function(eta,rho) {
  return(apply(eta,2,function(z){
    z.max <- max(z)
    return(ifelse(z.max>1-rho, which.max(z)[1], 0))
  }))
}

eta <- t(phat)

# choose cost of reject 'rho' so that total coverage is >= 0.9
tot.cov.eval <- function(rho) {
  res <- CWR(eta,rho)
  return( ( sum(res==Y2) + sum(res==0) ) / length(res) )
}

rho.grid <- seq(0,1-1/3,length.out=100)

tot.cov.grid <- sapply(rho.grid,tot.cov.eval)

# largest rho so that total coverage >= 0.9
(rho.optim <- rho.grid[ tail(which(tot.cov.grid >= 0.9),1) ])

# classifier with reject option with cost of reject = rho.optim, evaluated in test sample
res <- CWR(t(phat_test),rho.optim)

# ambiguity
mean( 3*(res == 0) + (res != 0) )
# total coverage
( sum(res==Y.test) + sum(res==0) ) / length(res)
# empirical class coverage
mean( (res==1 | res==0)[Y.test==1] )
mean( (res==2 | res==0)[Y.test==2] )
mean( (res==3 | res==0)[Y.test==3] )

