### Least Ambiguous Set-Valued Classifiers with Bounded Error Levels
### Example 5.2: A simple, 2-dimensional, simulated example using kernel and kNN classification
### Warning: plotting functions are taylored to this example, any changes will result in unexpected behaviour 

rm(list=ls())

##########################################################################################
##########################################################################################
########################              KERNEL              ################################
##########################################################################################


library(mvtnorm)

kernel = function(x,grid,h){
     ### kernel density estimator
     m = nrow(grid)
     p = rep(0,m)
     d = ncol(x)
     sigma = h^2*diag(d)
     for(i in 1:m){
          p[i] = mean(dmvnorm(x,grid[i,],sigma)) 
          }
     return(p)
}


bandwidth = function(X){
     ### Silverman's rule for bandwidth
     X = as.matrix(X)
     n = nrow(X)
     d = ncol(X)
     S = sqrt(mean(apply(X,2,var)))
     h = S*(4/(d+2))^(1/(d+4))*n^(-1/(d+4))
     return(h)
}



#######################################################
## fit model on (X1,Y1); predict on X2
#######################################################

fit <- function(X1,Y1,X2){
	 k = max(Y1)
	 n1 = length(Y1)
	 n2 = nrow(X2)
     pi = table(Y1)/n1
     P  = matrix(0,n2,k)
     for(j in 1:k){
          I = (1:n1)[Y1 == j]
          h = bandwidth(X1[I,])
          P[,j] = kernel(X1[I,],X2,h)
     }
     P = P*matrix(pi,n2,k,byrow=TRUE)
     s = apply(P,1,sum)
     phat = P/matrix(s,n2,k,byrow=FALSE) 
	 return(phat)
}

#######################################################
## Conformal thresholds for total coverage and class-specific coverage
#######################################################
conf_thlds <- function(phat,Y2,alpha_tot,class_alphas){
     m = nrow(phat)
	 K = ncol(phat)
     score = rep(0,m)
     for(i in 1:m){
          score[i] = phat[i,Y2[i]]
     }
	 ## Total coverage
	 tot_thlds = sort(score)[ceiling(alpha_tot*(m+1)-1)]
	 
	 ## Class-specific coverage
	 class_thlds <- rep(NA,K)
	 for (k in 1:K){
		class_thlds[k] <- sort(score[Y2==k])[ceiling(class_alphas[k]*(sum(Y2==k)+1)-1)]
	 }
	 
	 return(list(tot_thlds=tot_thlds,class_thlds=class_thlds))
}
	 
#######################################################
## plot prediction regions
#######################################################
plot_pred_regions <- function(xGrid,yGrid,PHAT,thld,dat){

     nx <- length(xGrid)
	 ny <- length(yGrid)
	 
	 plot(xGrid,yGrid,"n",xlim=c(-4,4),ylim=c(-2,5),xlab="",ylab="",
		 mar=c(1, 1, 4, 1) + 0.1,cex.axis=1.9,xaxt="n",yaxt="n")
	 contour1 <- contourLines(x = xGrid, y = yGrid, z = matrix(PHAT[,1],nx,ny), 
		 nlevels = 1, levels = thld[1])
	 contour2 <- contourLines(x = xGrid, y = yGrid, z = matrix(PHAT[,2],nx,ny), 
		 nlevels = 1, levels = thld[2])
	 contour3 <- contourLines(x = xGrid, y = yGrid, z = matrix(PHAT[,3],nx,ny), 
		 nlevels = 1, levels = thld[3])
	 
	 polygon(c(contour2[[1]]$x,-6,-6),c(contour2[[1]]$y,8,-6),col=rgb(0, 0, 1, 0.3),border=NA)
	 polygon(c(contour1[[1]]$x,8),c(contour1[[1]]$y,8),       col=rgb(1, 0, 0, 0.4),border=NA)
	 polygon(c(contour3[[1]]$x,8,8),c(contour3[[1]]$y,8,-6),  col=rgb(1, 1, 0, 0.4),border=NA)
	 
	 polygon(c(contour2[[1]]$x,-6,-6),c(contour2[[1]]$y,8,-6),density=10, angle=40, col="gray50",border="gray")
	 polygon(c(contour1[[1]]$x,8),c(contour1[[1]]$y,8),       density=10, angle=90,  col="gray50",border="gray")
	 polygon(c(contour3[[1]]$x,8,8),c(contour3[[1]]$y,8,-6),  density=10, angle=140, col="gray50",border="gray")
	 
	 class_cols <- c("red","blue","gold")
	 class_pch <- c(1,4,2)
	 points(dat$X,asp=1,pch=class_pch[dat$Y],col=class_cols[dat$Y],cex=1,xlab="",ylab="")
	 
	 text(-2,4.5,1,cex=2.8,font=4)
	 text(-3.8,-1.5,2,cex=2.8,font=4)
	 text(3.7,-1.5,3,cex=2.8,font=4)
}

plot_pred_regions_CWR <- function(xGrid,yGrid,PHAT,thld,dat){

     nx <- length(xGrid)
	 ny <- length(yGrid)
	 
	 plot(xGrid,yGrid,"n",xlim=c(-4,4),ylim=c(-2,5),xlab="",ylab="",
		 mar=c(1, 1, 4, 1) + 0.1,cex.axis=1.9,xaxt="n",yaxt="n")	 
	 contour1 <- contourLines(x = xGrid, y = yGrid, z = matrix(PHAT[,1],nx,ny), 
		 nlevels = 1, levels = thld[1])
	 contour2 <- contourLines(x = xGrid, y = yGrid, z = matrix(PHAT[,2],nx,ny), 
		 nlevels = 1, levels = thld[2])
	 contour3 <- contourLines(x = xGrid, y = yGrid, z = matrix(PHAT[,3],nx,ny), 
		 nlevels = 1, levels = thld[3])
	 
	 CWR2x <- c(contour3[[1]]$x,contour1[[1]]$x)
	 CWR2y <- c(contour3[[1]]$y,contour1[[1]]$y)
	 polygon(c(CWR2x,-8,-8),c(CWR2y,8,-8),col=rgb(0, 0, 1, 0.3),border=NA)
	 
	 CWR1x <- c(rev(contour2[[1]]$x),contour3[[1]]$x)
	 CWR1y <- c(rev(contour2[[1]]$y),contour3[[1]]$y)
	 polygon(c(CWR1x,8,-8),c(CWR1y,8,8),col=rgb(1, 0, 0, 0.4),border=NA)
	 
	 CWR3x <- c(contour1[[1]]$x,rev(contour2[[1]]$x))
	 CWR3y <- c(contour1[[1]]$y,rev(contour2[[1]]$y))
	 polygon(c(CWR3x,0,8,8),c(CWR3y,-8,-8,8),col=rgb(1, 1, 0, 0.4),border=NA)
	 
	 polygon(c(CWR2x,-8,-8),c(CWR2y,8,-8),density=10, angle=40, col="gray50",border="gray")
	 polygon(c(CWR1x,8,-8),c(CWR1y,8,8),       density=10, angle=90,  col="gray50",border="gray")
	 polygon(c(CWR3x,0,8,8),c(CWR3y,-8,-8,8),  density=10, angle=140, col="gray50",border="gray")
	 
	 class_cols <- c("red","blue","gold")
	 class_pch <- c(1,4,2)
	 points(dat$X,asp=1,pch=class_pch[dat$Y],col=class_cols[dat$Y],cex=1,xlab="",ylab="")
	 
     text(-2,4.5,1,cex=2.8,font=4)
	 text(-3.8,-1.5,2,cex=2.8,font=4)
	 text(3.7,-1.5,3,cex=2.8,font=4)
}

#######################################################
## Greedy expansion
#######################################################
# gridsize: maximum number of decrements for each threshold
greedy_fill <- function (phat,t_0,gridsize=100,seed=1234){
	# t_0: thresholds from original solution
	# we will decrease them greedily until they are <= 1
	n_class <- ncol(phat)
	
	potential_tholds <- lapply(t_0,seq,to=0,length.out=gridsize+2)
	potential_tholds <- lapply(potential_tholds,
								FUN=function(x) x[2:(gridsize+1)]) # thold cannot be 0
	i <- rep(0,n_class) # indicates number of decrements made to each threshold
	classes <- 1:n_class
	t_can_decrease <- t_0 > 0
	H_star <- t(apply(phat, 1, # current LABEL classifier in test sample
		function(x){
			as.numeric(x >= t_0)
		}))
	
	ambiguity <- mean(rowSums(H_star))
	# start greedy decrements
	t_iter <- t_0
	iter=1
	set.seed(seed)
	while(sum(t_iter)>1){
		iter=iter+1
		# potential new values of the tholds
		t_candidates <- mapply(function(x,y) x[y], potential_tholds, i+1)
		# if a thold cannot decrease then it is fixed at current value
		t_candidates[!t_can_decrease] <- t_iter[!t_can_decrease]
		
		H_candidates <-
			t(apply(phat, 1,
			function(x){
				as.numeric(x >= t_candidates)
			})) # this would be the new LABEL using t_candidates
		
		# these are the n_class possible values of ambiguity from changing one thold at a time
		amb_cand <- ambiguity[iter-1] - colMeans(H_star) + colMeans(H_candidates)
		amb_cand <- amb_cand[t_can_decrease]
		
		# pick the thold that increases ambiguity the least
		k_greedy <- which(amb_cand==min(amb_cand))
		k_greedy <- sample(x=k_greedy,size=1)
		k_greedy <- classes[t_can_decrease][k_greedy]
		
		t_iter[k_greedy] <- t_candidates[k_greedy]
		
		i[k_greedy] <- i[k_greedy] + 1
		
		t_can_decrease[k_greedy] <- i[k_greedy] < gridsize
			
		H_star[,k_greedy] <- H_candidates[,k_greedy]
		ambiguity[iter] <- mean(rowSums(H_star))
	}
	tholds <- t_iter
	return(tholds)
}

#######################################################
###### Example
###### Use kernel classification with split conformal
#######################################################

# Classes 2 and 3 overlap a lot 
set.seed(29531)
mean1 <- c(0,4) 
mean2 <- c(-2,0)
mean3 <- c(2,0)
means <- list(mean1, mean2, mean3)

covar1 <- diag(c(1,1)/3)
covar2 <- diag(c(2,2)/2)
covar3 <- diag(c(5,1)/2)
covars <- list(covar1, covar2, covar3)

means_covars <- list(list(mean1, covar1),
	list(mean2, covar2),
	list(mean3, covar3))

# Random sample
### Y takes values in {1,..,k}    

n <- rmultinom(1,2000,rep(1,3)) 
x1 <- rmvnorm(n=n[1], mean=mean1, sigma=covar1)
x2 <- rmvnorm(n=n[2], mean=mean2, sigma=covar2)
x3 <- rmvnorm(n=n[3], mean=mean3, sigma=covar3)
X <- rbind(x1,x2,x3)
Y <- c(rep(1,n[1]),rep(2,n[2]),rep(3,n[3]))
perm <- sample(nrow(X))
X <- X[perm,]
Y <- Y[perm]
#plot(X,col=c(rep("red",n[1]),rep("blue",n[2]),rep("yellow",n[3])))



#######################################################
###### Split the data for split-conformal prediction ##
#######################################################
n = length(Y)
m = round(n/2)
set.seed(29531)
I = sample(1:n,size=m,replace=FALSE)
### training data
Y1 = Y[I]
X1 = X[I,]
n1 = length(Y1)
## test data
J = setdiff(1:n,I)
Y2 = Y[J]
X2 = X[J,]
n2 = length(Y2)


#######################################################
## fit model on (X1,Y1); predict on X2
#######################################################
phat <- fit(X1,Y1,X2)

#######################################################
## compute the estimate p(y|x) over a grid (for plotting)
#######################################################
grid1  = seq(-7,7,length=100)
Grid  = as.matrix(expand.grid(grid1,grid1))
PHAT <- fit(X1,Y1,Grid)

#######################################################
## get thresholds for total and class specific coverage
#######################################################
thlds_2pc <- conf_thlds(phat,Y2,alpha_tot=.02,class_alphas=rep(.02,3))

# ambiguity of LABEL classifier with total coverage
H_star <- t(apply(phat, 1, 
		function(x){
			as.numeric(x >= rep(thlds_2pc$tot_thlds,3))
		}))	
mean(rowSums(H_star)) 



#######################################################
## Comparing to classification with rejection (CWR)
#######################################################

CWR <- function(eta,rho) {
  return(apply(eta,2,function(z){
    z.max <- max(z)
    return(ifelse(z.max>1-rho, which.max(z)[1], 0))
  }))
}

eta <- t(phat)

# choose cost of reject 'rho' so that total coverage is >= 0.98

tot.cov.eval <- function(rho) {
  res <- CWR(eta,rho)
  return( ( sum(res==Y2) + sum(res==0) ) / length(res) )
}

rho.grid <- seq(0,.5,length.out=100)

tot.cov.grid <- sapply(rho.grid,tot.cov.eval)

# largest rho so that total coverage >= 0.98
rho.optim <- rho.grid[ tail(which(tot.cov.grid >= 0.98),1) ]

# classifier with reject option with cost of reject = rho.optim
res <- CWR(eta,rho.optim)

# ambiguity
mean( 3*(res == 0) + (res != 0) )
# total coverage
( sum(res==Y2) + sum(res==0) ) / length(res)
# empirical class coverage
mean( (res==1 | res==0)[Y2==1] )
mean( (res==2 | res==0)[Y2==2] )
mean( (res==3 | res==0)[Y2==3] )


#######################################################
##### plot the prediction regions 
#######################################################

windows(15,7)
nf <- layout(matrix(c(11:15,9,1:4,10,5:8),3,5,byrow=TRUE), widths=c(1,rep(3.5,4)), heights=c(0.4,3,3), TRUE)
layout.show(nf)
par(mar = rep(.3,4))
axismarks <- c(-4,0,4)

plot_pred_regions(xGrid=grid1,yGrid=grid1,PHAT=PHAT,thld=rep(thlds_2pc$tot_thlds,3),dat=list(X=X2,Y=Y2))
mtext("LABEL Total Coverage = 0.98",side=3,font=2,line=1,cex=1)
mtext("Kernel Classifier",side=2,font=2,line=3.5,las=3,cex=1.2)
axis(side = 2, at = axismarks, cex.axis=2,las=0)
################# CWR plot
plot_pred_regions_CWR(xGrid=grid1,yGrid=grid1,PHAT=PHAT,thld=rep(1-rho.optim,3),dat=list(X=X2,Y=Y2))
mtext("CWR Total Coverage = 0.98",side=3,font=2,line=1,cex=1)
plot_pred_regions(xGrid=grid1,yGrid=grid1,PHAT=PHAT,thld=thlds_2pc$class_thlds,dat=list(X=X2,Y=Y2))
mtext("LABEL Class Coverage = 0.98",side=3,font=2,line=1,cex=1)
greedy_thlds_2pc <- greedy_fill(phat,t_0=thlds_2pc$class_thlds,gridsize=100,seed=1234)
plot_pred_regions(xGrid=grid1,yGrid=grid1,PHAT=PHAT,thld=greedy_thlds_2pc,dat=list(X=X2,Y=Y2))
mtext(expression(paste(bold("Accretive Completion, Class Cov."), bold(NA>=NA), bold("0.98"))),side=3,font=2,line=1,cex=.8)

# class coverage of classifier that controls for total coverage
round(sapply(1:3,function(k)mean(phat[Y2==k,k]>=thlds_2pc$tot_thlds)),3)

# ambiguity of final LABEL classifier
H_star <- t(apply(phat, 1, 
		function(x){
			as.numeric(x >= greedy_thlds_2pc)
		}))	
mean(rowSums(H_star)) 


##########################################################################################
##########################################################################################
########################              kNN                 ################################
##########################################################################################

#######################################################
## fit model on (X1,Y1); predict on X2
#######################################################

source("utils_V2.r") # file should be in working directory
library(FNN)
Y1.factor <- as.factor(Y1)
loo_cvs <- sapply(1:30,function(k) mean( knn.cv(X1, Y1.factor, k = k) == Y1.factor ) )
# plot(loo_cvs) # best predictive k here is 17, but it seems to be due to noise
# true best k seems to be between 13 and 17, so we take midpoint
k <- 15 

knn_labels_test <- KNN.Label(X1, Y1, k=k, X2) # each row is the label of k nn for a ranking subsample point
phat <- t(apply(knn_labels_test, 1,
            function(x){table(factor(x, levels = 1:3)) / k})) 

#######################################################
## compute the estimate p(y|x) over a grid (for plotting)
#######################################################
grid1  = seq(-7,7,length=100)
Grid  = as.matrix(expand.grid(grid1,grid1))
knn_labels_Grid <- KNN.Label(X1, Y1, k=k, Grid) # each row is the label of k nn for a ranking subsample point
PHAT <- t(apply(knn_labels_Grid, 1,
            function(x){table(factor(x, levels = 1:3)) / k})) 

#######################################################
## get thresholds for total and class specific coverage
#######################################################
thlds_2pc <- conf_thlds(phat,Y2,alpha_tot=.02,class_alphas=rep(.02,3))


#######################################################
## Comparing to classification with rejection (CWR)
#######################################################

CWR <- function(eta,rho) {
  return(apply(eta,2,function(z){
    z.max <- max(z)
    return(ifelse(z.max>1-rho, which.max(z)[1], 0))
  }))
}

eta <- t(phat)

# choose cost of reject 'rho' so that total coverage is >= 0.98

tot.cov.eval <- function(rho) {
  res <- CWR(eta,rho)
  return( ( sum(res==Y2) + sum(res==0) ) / length(res) )
}

rho.grid <- seq(0,.5,length.out=100)

tot.cov.grid <- sapply(rho.grid,tot.cov.eval)

# largest rho so that total coverage >= 0.98
rho.optim <- rho.grid[ tail(which(tot.cov.grid >= 0.98),1) ]

# classifier with reject option with cost of reject = rho.optim
res <- CWR(eta,rho.optim)

# ambiguity
mean( 3*(res == 0) + (res != 0) )
# total coverage
( sum(res==Y2) + sum(res==0) ) / length(res)
# empirical class coverage
mean( (res==1 | res==0)[Y2==1] )
mean( (res==2 | res==0)[Y2==2] )
mean( (res==3 | res==0)[Y2==3] )

#######################################################
##### plot the prediction regions 
#######################################################

plot_pred_regions(xGrid=grid1,yGrid=grid1,PHAT=PHAT,thld=rep(thlds_2pc$tot_thlds,3),dat=list(X=X2,Y=Y2))
mtext("Nearest Neighbors",side=2,font=2,line=3.5,las=3,cex=1.2)
axis(side = 2, at = axismarks, cex.axis=2,las=0)
axis(side = 1, at = axismarks, cex.axis=2,las=0)
################# CWR plot
plot_pred_regions_CWR(xGrid=grid1,yGrid=grid1,PHAT=PHAT,thld=rep(1-rho.optim,3),dat=list(X=X2,Y=Y2))
axis(side = 1, at = axismarks, cex.axis=2,las=0)
plot_pred_regions(xGrid=grid1,yGrid=grid1,PHAT=PHAT,thld=thlds_2pc$class_thlds,dat=list(X=X2,Y=Y2))
axis(side = 1, at = axismarks, cex.axis=2,las=0)
greedy_thlds_2pc <- greedy_fill(phat,t_0=thlds_2pc$class_thlds,gridsize=100,seed=1234)
plot_pred_regions(xGrid=grid1,yGrid=grid1,PHAT=PHAT,thld=greedy_thlds_2pc,dat=list(X=X2,Y=Y2))
axis(side = 1, at = axismarks, cex.axis=2,las=0)
# class coverage:
# when fixing total coverage
round(sapply(1:3,function(k)mean(phat[Y2==k,k]>=thlds_2pc$tot_thlds)),3)
# when fixing class coverage
round(sapply(1:3,function(k)mean(phat[Y2==k,k]>=thlds_2pc$class_thlds[k])),3)

# ambiguity of final LABEL classifier
H_star <- t(apply(phat, 1, 
		function(x){
			as.numeric(x >= greedy_thlds_2pc)
		}))	
mean(rowSums(H_star)) 

