### Least Ambiguous Set-Valued Classifiers with Bounded Error Levels
### Example 5.3: Iris dataset

#######################################################
## LABEL thresholds for total coverage and class-specific coverage
#######################################################
thlds <- function(phat,Y2,alpha_tot,class_alphas){
     m = nrow(phat)
	 k = ncol(phat)
     score = rep(0,m)
     for(i in 1:m){
          score[i] = phat[i,Y2[i]]
     }
	 ## Total coverage
     tot_thlds = quantile(score,alpha_tot)

	 ## Class-specific coverage
	 class_thlds <- rep(NA,k)
	 for (i in 1:k){
		class_thlds[i] <- quantile(phat[Y2==i,i],probs=class_alphas[i])
	 }
	 
	 return(list(tot_thlds=tot_thlds,class_thlds=class_thlds))
}

#######################################################
## Accretive completion
#######################################################
# gridsize: maximum number of decrements for each threshold
accretive <- function (phat,t_0,gridsize=100,seed=1234){
	# t_0: thresholds from original solution
	# we will decrease them greedily until they are <= 1
	n_class <- ncol(phat)
	
	potential_tholds <- lapply(t_0,seq,to=0,length.out=gridsize+2)
	potential_tholds <- lapply(potential_tholds,
								FUN=function(x) x[2:(gridsize+1)]) # thold cannot be 0
	i <- rep(0,n_class) # indicates number of decrements made to each threshold
	classes <- 1:n_class
	t_can_decrease <- t_0 > 0
	H_star <- t(apply(phat, 1, # current LABEL classifier in test sample
		function(x){
			as.numeric(x >= t_0)
		}))
	
	ambiguity <- mean(rowSums(H_star))
	# start greedy decrements
	t_iter <- t_0
	iter=1
	set.seed(seed)
	while(sum(t_iter)>1){
		iter=iter+1
		# potential new values of the tholds
		t_candidates <- mapply(function(x,y) x[y], potential_tholds, i+1)
		# if a thold cannot decrease then it is fixed at current value
		t_candidates[!t_can_decrease] <- t_iter[!t_can_decrease]
		
		H_candidates <-
			t(apply(phat, 1,
			function(x){
				as.numeric(x >= t_candidates)
			})) # this would be the new LABEL using t_candidates
		
		# these are the n_class possible values of ambiguity from changing one thold at a time
		amb_cand <- ambiguity[iter-1] - colMeans(H_star) + colMeans(H_candidates)
		amb_cand <- amb_cand[t_can_decrease]
		
		# pick the thold that increases ambiguity the least
		k_greedy <- which(amb_cand==min(amb_cand))
		k_greedy <- sample(x=k_greedy,size=1)
		k_greedy <- classes[t_can_decrease][k_greedy]
		
		t_iter[k_greedy] <- t_candidates[k_greedy]
		
		i[k_greedy] <- i[k_greedy] + 1
		
		t_can_decrease[k_greedy] <- i[k_greedy] < gridsize
			
		H_star[,k_greedy] <- H_candidates[,k_greedy]
		ambiguity[iter] <- mean(rowSums(H_star))
	}
	tholds <- t_iter
	return(tholds)
}


## Iris data
data(iris)
set.seed(1955)
X = iris[,-5]
z = iris[,5]
n = length(z)
Y = rep(0,n)
Y[z == "setosa"]     = 1
Y[z == "versicolor"] = 2
Y[z == "virginica"]  = 3
X = as.matrix(X)
Iris <- data.frame(Y=as.factor(Y),X)

library(nnet)
fitted.model <- multinom(Y~.,data=Iris)

phat <- fitted(fitted.model)

#######################################################
## get thresholds for total and class specific coverage
#######################################################
thlds_est <- thlds(phat,Y,alpha_tot=.02,class_alphas=rep(.02,3))

Hstar_classcov <- t(apply(phat, 1, 
		function(x){
			as.numeric(x >= thlds_est$class_thlds)
		}))	
mean(rowSums(Hstar_classcov)) 

accr_thlds <- accretive(phat,t_0=thlds_est$class_thlds,gridsize=100,seed=1234)
Hstar_accr <- t(apply(phat, 1, 
		function(x){
			as.numeric(x >= accr_thlds)
		}))	
mean(rowSums(Hstar_accr)) # ambiguity

mean(c(Hstar_accr[Y==1,1],Hstar_accr[Y==2,2],Hstar_accr[Y==3,3])) # total coverage


ambi.ind <- which(apply(Hstar_accr,1,sum)>1)

X.pca <- princomp(X)
Y.color <- rep('red',n)
Y.color[which(Y==2)] <- 'gold'
Y.color[which(Y==3)] <- 'blue'


windows(8,6)
par(mar = c(4.5,4.5,3,1))
plot(X.pca$score[,1], X.pca$score[,2],col=Y.color,pch=as.character(Y),xlab='Principal Component 1',ylab='Principal Component 2',
	cex.axis=1.5,cex.lab=1.8,lwd=3,cex=1.5,main="Iris Data",cex.main=2)
points(X.pca$score[ambi.ind,1],X.pca$score[ambi.ind,2],col='forestgreen',pch=0,lwd=4,cex=2.5)


#######################################################
## Comparing to classification with rejection (CWR)
#######################################################

CWR <- function(eta,rho) {
  return(apply(eta,2,function(z){
    z.max <- max(z)
    return(ifelse(z.max>1-rho, which.max(z)[1], 0))
  }))
}

eta <- t(phat)

# choose cost of reject 'rho' so that total coverage is >= 0.98

tot.cov.eval <- function(rho) {
  res <- CWR(eta,rho)
  return( ( sum(res==Y) + sum(res==0) ) / length(res) )
}

rho.grid <- seq(0,1-1/3,length.out=100)

tot.cov.grid <- sapply(rho.grid,tot.cov.eval)

# largest rho so that total coverage >= 0.98
rho.optim <- rho.grid[ tail(which(tot.cov.grid >= 0.98),1) ]

# classifier with reject option with cost of reject = rho.optim
res <- CWR(eta,rho.optim)

# ambiguity
mean( 3*(res == 0) + (res != 0) )
# total coverage
( sum(res==Y) + sum(res==0) ) / length(res)
# empirical class coverage
mean( (res==1 | res==0)[Y==1] )
mean( (res==2 | res==0)[Y==2] )
mean( (res==3 | res==0)[Y==3] )

