# Simulation study to compare LABEL classifiers with classifiers with reject option
# in a multivariate scenario.

# We use the Abalone data to create a population from which we simulate. 
# We start by creating K=5 classes using the age variable.  
# We then use the numeric features of the Abalone data within each class 
# to obtain a mean and a covariance matrix.
# The population is then defined as a 5-component mixture of 7-variate normal distributions, 
# with means, covariance matrices, and population proportions obtained from the abalone data.  

# Each of the 1000 simulation replicates consists in n = 4000 draws from this mixture. 

# The estimator of p(y|x) is a simple multinomial logistic regression.

# For LABEL we used total coverage of 0.95. 
# For classifiers with reject option we chose cost of reject 'rho' so that total coverage is >= 0.95

# All coverage and ambiguity quantities are evaluated in-sample.  This is ok since the estimator is simple enough.

rm(list = ls())

# LABEL thresholds for total coverage and class-specific coverage
thlds <- function(phat,Y2,alpha_tot,class_alphas){
     m = nrow(phat)
	 k = ncol(phat)
     score = rep(0,m)
     for(i in 1:m){
          score[i] = phat[i,Y2[i]]
     }
	 ## Total coverage
     tot_thlds = quantile(score,alpha_tot)

	 ## Class-specific coverage
	 class_thlds <- rep(NA,k)
	 for (i in 1:k){
		class_thlds[i] <- quantile(phat[Y2==i,i],probs=class_alphas[i])
	 }
	 
	 return(list(tot_thlds=tot_thlds,class_thlds=class_thlds))
}

CWR <- function(eta,rho) {
	return(apply(eta,2,function(z){
		z.max <- max(z)
		return(ifelse(z.max>1-rho, which.max(z)[1], 0))
	}))
}

tot.cov.eval <- function(rho) {
	res <- CWR(eta,rho)
	return( ( sum(res==Y) + sum(res==0) ) / length(res) )
}

### Simulation set up
# Create population based on the Abalone data
library(AppliedPredictiveModeling)
library(MASS)
library(nnet)

data(abalone)
Y <- abalone[,9]
Y[Y>20] <- 20
Y <- Y%/%5 + 1
K <- max(Y)
X <- abalone[,2:8]
class_means <- sapply(1:K, function(k) colMeans(X[Y==k,]))
class_vars <- lapply(1:K, function(k) var(X[Y==k,]))
class_probs <- table(Y)/length(Y)

set.seed(59872)
simrep <- 1000
n <- 4000

LABEL_ambiguity <- rep(NA,simrep)
LABEL_class_coverage <- matrix(NA,simrep,K)
CWR_ambiguity <- rep(NA,simrep)
CWR_class_coverage <- matrix(NA,simrep,K)

system.time(
for(r in 1:simrep){
	ns <- rmultinom(1,n,class_probs)
	Y  <- rep(1:K,times=ns)
	X <- lapply(1:K,function(k)mvrnorm(n = ns[k], mu=class_means[,k], Sigma=class_vars[[k]]))
	X <- do.call(rbind,X)
	dat <- data.frame(Y=as.factor(Y),X)
	fitted.model <- multinom(Y~.,data=dat)
	phat <- fitted(fitted.model)
	
	# get thresholds for total coverage
	thlds_est <- thlds(phat,Y,alpha_tot=.05,class_alphas=rep(.05,K))
	
	Hstar_totcov <- t(apply(phat, 1, 
		function(x){
			as.numeric(x >= rep(thlds_est$tot_thlds,K))
		}))	
	
	LABEL_ambiguity[r] <- mean(rowSums(Hstar_totcov))
	LABEL_class_coverage[r,] <- sapply(1:K,function(k)mean(Hstar_totcov[Y==k,k]))
	
	### Classification with reject option	
	eta <- t(phat)
	
	# choose cost of reject 'rho' so that total coverage is >= 0.95
	rho.grid <- seq(0,1-1/K,length.out=100)
	tot.cov.grid <- sapply(rho.grid,tot.cov.eval)
	
	# largest rho so that total coverage >= 0.95
	rho.optim <- rho.grid[ tail(which(tot.cov.grid >= 0.95),1) ]
	
	# classifier with reject option with cost of reject = rho.optim
	res <- CWR(eta,rho.optim)
	
	CWR_ambiguity[r] <- mean( K*(res == 0) + (res != 0) )
	CWR_class_coverage[r,] <- sapply(1:K,function(k) mean( (res==k | res==0)[Y==k] ))
	
}
)

# Average class-coverage
colMeans(LABEL_class_coverage)
colMeans(CWR_class_coverage)

# Average ambiguity
mean(LABEL_ambiguity)
mean(CWR_ambiguity)
plot(LABEL_ambiguity,CWR_ambiguity,xlim=c(1.8,3.9),ylim=c(1.8,3.9),
xlab="LABEL",ylab="CWR",main="Ambiguity")
abline(0,1,col="red")

# percentage of times LABEL ambiguity is larger than CWR ambiguity
# in the three simulation scenarios
mean(LABEL_ambiguity>CWR_ambiguity)

