# Simulation study to compare LABEL classifiers with classifiers with reject option
# in a univariate scenario.

# We simulate n = 4000 draws from a mixture with K = 3 components.
# with distributions N(-2,1), N(0,1), N(2,1).  We explore three scenarios of 
# the population class probabilities: (.45,.1,.45), (0.33, 0.33, 0.34), (0.6, 0.3, 0.1)

# The density estimator is a Gaussian density.  That is, using the labeled data, 
# we estimate the Gaussian parameters and then evaluate Gaussian densities to get p(y|x).

# For LABEL we used total coverage of 0.95. 
# For classifiers with reject option we chose cost of reject 'rho' so that total coverage is >= 0.95

# All coverage and ambiguity quantities are evaluated in-sample.  This is ok since the estimator is simple enough.
# We repeat the simulation 1000 times and report average results.

### Simulation set up
rm(list = ls())
set.seed(59872)
simrep <- 1000
n <- 4000
pop.class.probs <- list( c(.45,.1,.45), c(0.33, 0.33, 0.34), c(0.6, 0.3, 0.1) )
as <- c(0.05,0.05,0.05)
a.tot.cov <- 0.05

LABEL_ambiguity <- list(rep(NA,simrep),rep(NA,simrep),rep(NA,simrep))
LABEL_class_coverage <- list(matrix(NA,simrep,3),matrix(NA,simrep,3),matrix(NA,simrep,3))
CWR_ambiguity <- list(rep(NA,simrep),rep(NA,simrep),rep(NA,simrep))
CWR_class_coverage <- list(matrix(NA,simrep,3),matrix(NA,simrep,3),matrix(NA,simrep,3))

EtaFunc <- function(x.vals,m1,s1,p1,m2,s2,p2,m3,s3,p3) {
  jp1.grid <- dnorm(x.vals, m1, s1) * p1
  jp2.grid <- dnorm(x.vals, m2, s2) * p2
  jp3.grid <- dnorm(x.vals, m3, s3) * p3
  mpx.grid <- jp1.grid + jp2.grid + jp3.grid
  eta1 <- jp1.grid / mpx.grid
  eta2 <- jp2.grid / mpx.grid
  eta3 <- 1 - eta1 - eta2
  return(rbind(eta1, eta2, eta3))
}

ConfSet <- function(x.grid,t1,t2,t3,m1,s1,p1,m2,s2,p2,m3,s3,p3) {
  eta <- EtaFunc(x.grid,m1,s1,p1,m2,s2,p2,m3,s3,p3)
  return(data.frame(c1=eta[1,]>t1,c2=eta[2,]>t2,c3=eta[3,]>t3))
}

ConfSet.tot.cov <- function(x.grid,thld,m1,s1,p1,m2,s2,p2,m3,s3,p3) {
eta <- EtaFunc(x.grid,m1,s1,p1,m2,s2,p2,m3,s3,p3)
return(data.frame(c1=eta[1,]>thld,c2=eta[2,]>thld,c3=eta[3,]>thld))
}

CWR <- function(eta,rho) {
	return(apply(eta,2,function(z){
		z.max <- max(z)
		return(ifelse(z.max>1-rho, which.max(z)[1], 0))
	}))
}

tot.cov.eval <- function(rho) {
	res <- CWR(eta,rho)
	return( ( sum(res==y) + sum(res==0) ) / n )
}

system.time(
for(v in 1:3){
for(r in 1:simrep){
	ns <- rmultinom(1,n,pop.class.probs[[v]]) 
	x1 <- -2+rnorm(ns[1])
	y1 <- rep(1,ns[1])
	x2 <- rnorm(ns[2])
	y2 <- rep(2,ns[2])
	x3 <- 2+rnorm(ns[3])
	y3 <- rep(3,ns[3])
	
	x <- c(x1,x2,x3)
	y <- c(y1,y2,y3)
	
	m1 <- mean(x1)
	s1 <- sd(x1)
	m2 <- mean(x2)
	s2 <- sd(x2)
	m3 <- mean(x3)
	s3 <- sd(x3)
	p1 <- ns[1] / n
	p2 <- ns[2] / n
	p3 <- 1-p1-p2
	
	### Classification with confidence -- class coverage
	eta1.x1 <- EtaFunc(x1,m1,s1,p1,m2,s2,p2,m3,s3,p3)[1,]
	#t1 <- sort(eta1.x1)[floor(ns[1]*as[1])]
	
	eta2.x2 <- EtaFunc(x2,m1,s1,p1,m2,s2,p2,m3,s3,p3)[2,]
	#t2 <- sort(eta2.x2)[floor(ns[2]*as[2])]
	
	eta3.x3 <- EtaFunc(x3,m1,s1,p1,m2,s2,p2,m3,s3,p3)[3,]
	#t3 <- sort(eta3.x3)[floor(ns[3]*as[3])]
	
	# which labels are assigned to the x points?
	#cs <- ConfSet(x,t1,t2,t3,m1,s1,p1,m2,s2,p2,m3,s3,p3)
	
	# ambiguity
	#mean(rowSums(cs))
	# total coverage
	#mean(c( cs[1:ns[1],1], cs[(ns[1]+1):(ns[1]+ns[2]),2], cs[(ns[1]+ns[2]+1):n,3] ))
	# empirical class coverage
	#mean(cs[1:ns[1],1])
	#mean(cs[(ns[1]+1):(ns[1]+ns[2]),2])
	#mean(cs[(ns[1]+ns[2]+1):n,3])
	
	# any unlabeled points?
	#any(rowSums(cs)==0)
	# are labels 1 and 3 ever assigned to the same point?
	#any(cs[,1]&cs[,3])
	
	
	### LABEL Classification -- total coverage
	eta.comb <- c(eta1.x1,eta2.x2,eta3.x3)
	t.tot.cov <- sort(eta.comb)[floor(n*a.tot.cov)]
		
	# which labels are assigned to the x points?
	cs.tot.cov <- ConfSet.tot.cov(x,t.tot.cov,m1,s1,p1,m2,s2,p2,m3,s3,p3)
	
	# ambiguity
	LABEL_ambiguity[[v]][r] <- mean(rowSums(cs.tot.cov))
	# total coverage
	#mean(c( cs.tot.cov[1:ns[1],1], cs.tot.cov[(ns[1]+1):(ns[1]+ns[2]),2], cs.tot.cov[(ns[1]+ns[2]+1):n,3] ))
	# empirical class coverage
	LABEL_class_coverage[[v]][r,1] <- mean(cs.tot.cov [1:ns[1],1])
	LABEL_class_coverage[[v]][r,2] <- mean(cs.tot.cov [(ns[1]+1):(ns[1]+ns[2]),2])
	LABEL_class_coverage[[v]][r,3] <- mean(cs.tot.cov [(ns[1]+ns[2]+1):n,3])
	
	# any unlabeled points?
	#any(rowSums(cs.tot.cov)==0)
	# are labels 1 and 3 ever assigned to the same point?
	#any(cs.tot.cov[,1]&cs.tot.cov[,3])
	
	
	### Classification with reject option
		
	eta <- EtaFunc(x,m1,s1,p1,m2,s2,p2,m3,s3,p3)
	
	# choose cost of reject 'rho' so that total coverage is >= 1 - a.tot.cov
		
	rho.grid <- seq(0,0.5,length.out=100)
	
	tot.cov.grid <- sapply(rho.grid,tot.cov.eval)
	
	# largest rho so that total coverage >= 1 - a.tot.cov
	rho.optim <- rho.grid[ tail(which(tot.cov.grid >= 1 - a.tot.cov),1) ]
	
	# classifier with reject option with cost of reject = rho.optim
	res <- CWR(eta,rho.optim)
	
	# ambiguity
	CWR_ambiguity[[v]][r] <- mean( 3*(res == 0) + (res != 0) )
	# total coverage
	#( sum(res==y) + sum(res==0) ) / n
	# empirical class coverage
	CWR_class_coverage[[v]][r,1] <- mean( (res==1 | res==0)[1:ns[1]] )
	CWR_class_coverage[[v]][r,2] <- mean( (res==2 | res==0)[(ns[1]+1):(ns[1]+ns[2])] )
	CWR_class_coverage[[v]][r,3] <- mean( (res==3 | res==0)[(ns[1]+ns[2]+1):n] )	
}}
)

# Average class-coverage
lapply(LABEL_class_coverage,FUN=function(x)round(colMeans(x),2))
lapply(CWR_class_coverage,FUN=function(x)round(colMeans(x),2))

# Average ambiguity
round(sapply(LABEL_ambiguity,FUN=mean),2)
round(sapply(CWR_ambiguity,FUN=mean),2)

# percentage of times LABEL ambiguity is larger than CWR ambiguity
# in the three simulation scenarios
colMeans(sapply(1:3,function(v)LABEL_ambiguity[[v]]>CWR_ambiguity[[v]]))

