### Least Ambiguous Set-Valued Classifiers with Bounded Error Levels
### Example 5.5: Zip code dataset

rm(list = ls())
library(FNN)
library(nnet)
library(caret)

#######################################################
## Conformal LABEL thresholds for total coverage and class-specific coverage
#######################################################
conf_thlds <- function(phat,Y2,alpha_tot,class_alphas){
     m = nrow(phat)
	 K = ncol(phat)
     score = rep(0,m)
     for(i in 1:m){
          score[i] = phat[i,Y2[i]]
     }
	 ## Total coverage
	 tot_thlds = sort(score)[ceiling(alpha_tot*(m+1)-1)]
	 
	 ## Class-specific coverage
	 class_thlds <- rep(NA,K)
	 for (k in 1:K){
		class_thlds[k] <- sort(score[Y2==k])[ceiling(class_alphas[k]*(sum(Y2==k)+1)-1)]
	 }
	 
	 return(list(tot_thlds=tot_thlds,class_thlds=class_thlds))
}

#######################################################
## Greedy expansion
#######################################################
# gridsize: maximum number of decrements for each threshold
greedy_fill <- function (phat,t_0,gridsize=100,seed=1234){
	# t_0: thresholds from original solution
	# we will decrease them greedily until they are <= 1
	n_class <- ncol(phat)
	
	potential_tholds <- lapply(t_0,seq,to=0,length.out=gridsize+2)
	potential_tholds <- lapply(potential_tholds,
								FUN=function(x) x[2:(gridsize+1)]) # thold cannot be 0
	i <- rep(0,n_class) # indicates number of decrements made to each threshold
	classes <- 1:n_class
	t_can_decrease <- t_0 > 0
	H_star <- t(apply(phat, 1, # current LABEL classifier in test sample
		function(x){
			as.numeric(x >= t_0)
		}))
	
	ambiguity <- mean(rowSums(H_star))
	# start greedy decrements
	t_iter <- t_0
	iter=1
	set.seed(seed)
	while(sum(t_iter)>1){
		iter=iter+1
		# potential new values of the tholds
		t_candidates <- mapply(function(x,y) x[y], potential_tholds, i+1)
		# if a thold cannot decrease then it is fixed at current value
		t_candidates[!t_can_decrease] <- t_iter[!t_can_decrease]
		
		H_candidates <-
			t(apply(phat, 1,
			function(x){
				as.numeric(x >= t_candidates)
			})) # this would be the new LABEL using t_candidates
		
		# these are the n_class possible values of ambiguity from changing one thold at a time
		amb_cand <- ambiguity[iter-1] - colMeans(H_star) + colMeans(H_candidates)
		amb_cand <- amb_cand[t_can_decrease]
		
		# pick the thold that increases ambiguity the least
		k_greedy <- which(amb_cand==min(amb_cand))
		k_greedy <- sample(x=k_greedy,size=1)
		k_greedy <- classes[t_can_decrease][k_greedy]
		
		t_iter[k_greedy] <- t_candidates[k_greedy]
		
		i[k_greedy] <- i[k_greedy] + 1
		
		t_can_decrease[k_greedy] <- i[k_greedy] < gridsize
			
		H_star[,k_greedy] <- H_candidates[,k_greedy]
		ambiguity[iter] <- mean(rowSums(H_star))
	}
	tholds <- t_iter
	return(tholds)
}





### Zip code data
load('zip_permute.RData')
Y.train.p <- Y.train.p + 1 # each digit is represented as the digit + 1. "0" is used as reject option in CWR
Y.test.p  <- Y.test.p  + 1 

# Split the training data for split-conformal prediction
n.train = length(Y.train.p)
m = round(n.train/3)
I = sample(1:n.train, size=2*m, replace=FALSE)
# model fitting data
Y1 = Y.train.p[I]
X1 = X.train.p[I,]
n1 = length(Y1)
# threshold estimation data
J = setdiff(1:n.train,I)
Y2 = Y.train.p[J]
X2 = X.train.p[J,]
n2 = length(Y2)


### split conformal
library(FNN)
source("utils_V2.r")
k=10
knn_labels <- KNN.Label(X1, Y1, k=k, X2) 
phat <- t(apply(knn_labels, 1,
            function(x){table(factor(x, levels = 1:10)) / k})) 
# get thresholds for total and class specific coverage
(thlds_est <- conf_thlds(phat,Y2,alpha_tot=.05,class_alphas=rep(.05,10)))

# Evaluate ambiguity and coverage on test data
knn_labels_test <- KNN.Label(X1, Y1, k=k, X.test.p) 
phat_test <- t(apply(knn_labels_test, 1,
            function(x){table(factor(x, levels = 1:10)) / k})) 

# test sample ambiguity of LABEL classifier with class coverage
H_star_test <- t(apply(phat_test, 1, 
		function(x){
			as.numeric(x >= thlds_est$class_thlds)
		}))	
nlab_test <- rowSums(H_star_test)
mean(nlab_test) # test sample ambiguity
sum(nlab_test==0) # test instances in null region
test_ambi_init <- nlab_test > 1 # ambiguous instances 


# ambiguous images between 3, 5
tmp.ind <- which(H_star_test[,4] + H_star_test[,6] > 1)
par(mfrow=c(10,10),mar = c(0,0,0,0))
for (i in 1:length(tmp.ind)) {
  image(z = FlipMatrix(matrix(as.numeric(X.test.p[tmp.ind[i], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))
}
plot.ind <- c(2,22,25)
pdf('bayes_35.pdf',width=6,height=2)
par(mfrow=c(1,3),mar=c(1,1,1,1))
for (i in 1:3) {
  image(FlipMatrix(matrix(as.numeric(X.test.p[tmp.ind[plot.ind[i]], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))  
}
dev.off()

# ambiguous images between 4, 9
tmp.ind <- which(H_star_test[,5] + H_star_test[,10] > 1)
par(mfrow=c(10,10),mar = c(0,0,0,0))
for (i in 1:length(tmp.ind)) {
  image(z = FlipMatrix(matrix(as.numeric(X.test.p[tmp.ind[i], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))
}
plot.ind <- c(4,6,15)
pdf('bayes_49.pdf',width=6,height=2)
par(mfrow=c(1,3),mar=c(1,1,1,1))
for (i in 1:3) {
  image(FlipMatrix(matrix(as.numeric(X.test.p[tmp.ind[plot.ind[i]], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))  
}
dev.off()

# ambiguous images between 7, 8, 9
tmp.ind <- which(rowSums(H_star_test[,8:10])==3)
par(mfrow=c(10,10),mar = c(0,0,0,0))
for (i in 1:length(tmp.ind)) {
  image(z = FlipMatrix(matrix(as.numeric(X.test.p[tmp.ind[i], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))
}
plot.ind <- c(1,2,3)
pdf('bayes_789.pdf',width=6,height=2)
par(mfrow=c(1,3),mar=c(1,1,1,1))
for (i in 1:3) {
  image(FlipMatrix(matrix(as.numeric(X.test.p[tmp.ind[plot.ind[i]], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))  
}
dev.off()


### Now use accretive completion to eliminate null region
greedy_thlds <- greedy_fill(phat,t_0=thlds_est$class_thlds,gridsize=100,seed=1234)

# LABEL classifier with accretive thlds
H_star2_test <- t(apply(phat_test, 1, 
		function(x){
			as.numeric(x >= greedy_thlds)
		}))	
nlab2_test <- rowSums(H_star2_test)
mean(nlab2_test) # test sample ambiguity
test_ambi2 <- nlab2_test > 1 # ambiguous instances 
mean(H_star2_test[cbind(1:length(Y.test.p),Y.test.p[1:length(Y.test.p)])]) # test sample coverage
table(nlab2_test)

# Co-occurrence matrix
t(H_star2_test) %*% H_star2_test

# images reported ambiguous after accretive completion
new_ambi <- which(test_ambi2 & (nlab_test<2))
par(mfrow=c(25,10),mar = c(0,0,0,0))
for (i in 1:length(new_ambi)) {
  image(z = FlipMatrix(matrix(as.numeric(X.test.p[new_ambi[i], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))
}


# additional ambiguous images between 3 and 5 
tmp.ind <- which(rowSums(H_star2_test[new_ambi,c(4,6)])==2)
par(mfrow=c(7,10),mar = c(0,0,0,0))
for (i in 1:length(tmp.ind)) {
  image(z = FlipMatrix(matrix(as.numeric(X.test.p[new_ambi[tmp.ind][i], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))
}
plot.ind <- c(15,20,30)
pdf('accr_35.pdf',width=6,height=2)
par(mfrow=c(1,3),mar=c(1,1,1,1))
for (i in 1:3) {
  image(FlipMatrix(matrix(as.numeric(X.test.p[new_ambi[tmp.ind][plot.ind[i]], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))  
}
dev.off()

# additional ambiguous images between 1 and 4
tmp.ind <- which(rowSums(H_star2_test[new_ambi,c(2,5)])==2)
par(mfrow=c(7,5),mar = c(0,0,0,0))
for (i in 1:length(tmp.ind)) {
  image(z = FlipMatrix(matrix(as.numeric(X.test.p[new_ambi[tmp.ind][i], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))
}
plot.ind <- c(2,3,19)
pdf('accr_14.pdf',width=6,height=2)
par(mfrow=c(1,3),mar=c(1,1,1,1))
for (i in 1:3) {
  image(FlipMatrix(matrix(as.numeric(X.test.p[new_ambi[tmp.ind][plot.ind[i]], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))  
}
dev.off()

# additional ambiguous images between 0, 2 and 3
tmp.ind <- which(rowSums(H_star2_test[new_ambi,c(1,3,4)])==3)
par(mfrow=c(7,5),mar = c(0,0,0,0))
for (i in 1:length(tmp.ind)) {
  image(z = FlipMatrix(matrix(as.numeric(X.test.p[new_ambi[tmp.ind][i], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))
}
plot.ind <- c(1,4,6)
pdf('accr_023.pdf',width=6,height=2)
par(mfrow=c(1,3),mar=c(1,1,1,1))
for (i in 1:3) {
  image(FlipMatrix(matrix(as.numeric(X.test.p[new_ambi[tmp.ind][plot.ind[i]], ]), ncol = 16)),
        xaxt = 'n', yaxt = 'n', col = gray(0:255/255))  
}
dev.off()



#######################################################
## Comparing to classification with rejection (CWR)
#######################################################

CWR <- function(eta,rho) {
  return(apply(eta,2,function(z){
    z.max <- max(z)
    return(ifelse(z.max>1-rho, which.max(z)[1], 0))
  }))
}

eta <- t(phat)

# choose cost of reject 'rho' so that total coverage is >= 0.98
tot.cov.eval <- function(rho) {
  res <- CWR(eta,rho)
  return( ( sum(res==Y2) + sum(res==0) ) / length(res) )
}

rho.grid <- seq(0,1-1/10,length.out=100)

tot.cov.grid <- sapply(rho.grid,tot.cov.eval)

# largest rho so that total coverage >= 0.98
rho.optim <- rho.grid[ tail(which(tot.cov.grid >= 0.98),1) ]

# classifier with reject option with cost of reject = rho.optim, evaluated in test sample
res <- CWR(t(phat_test),rho.optim)

# ambiguity
mean( 10*(res == 0) + (res != 0) )
# total coverage
( sum(res==Y.test.p) + sum(res==0) ) / length(res)
# empirical class coverage
sapply(1:10,function(cl) mean( (res==cl | res==0)[Y.test.p==cl] ))

